c$Id:$
      subroutine usolve(flags,b)

c      * * F E A P * * A Finite Element Analysis Program

c....  Copyright (c) 1984-2015: Regents of the University of California
c                               All rights reserved

c-----[--.----+----.----+----.-----------------------------------------]
c     Purpose:  Solver interface for UMFPACK

c     Inputs:
c        flags(1) - Allocation and/or initialization phase
c        flags(2) - Perform factorization for direct solutions
c        flags(3) - Coefficient array unsymmetric
c        flags(4) - Solve equations
c        flags(5) - Purge storage of pointers
c        b(*)     - RHS vector

c     Outputs:

c        flags(5) - True if error occurs (for factor/solve only)
c-----[--.----+----.----+----.-----------------------------------------]
      implicit   none

      include   'cdata.h'
      include   'compac.h'
      include   'debugs.h'
      include   'endata.h'
      include   'eqsym.h'
      include   'iofile.h'
      include   'part0.h'
      include   'setups.h'
      include   'pointer.h'
      include   'comblk.h'

      character  tname*5
      logical    flags(*), setvar,palloc
      integer    i, nnu, sys
      integer*8  symbolic, numeric                  ! For 64 bit pointers
c     integer    symbolic, numeric                  ! For 32 bit pointers
      real*8     b(*), dot, control(20), info(90)

      save

c     Presolve setups

      if(flags(1)) then

c       Assign storage type for unsymmetric forms

        uall   = .true.       ! Assemble total array
        ubycol = .true.       ! Store by columns
        udiag  = .true.       ! Diagonal in sparse array
        ulfl   = .false.      ! Prevent double assembly in cassem

c       Allow for assembly of global equations to tangent matrix

        gceflg = .true.

c       Establish current matrix indices

        call iters(nnu,-1)    ! Forces call to uiters

c       Allocate matrix storage

        write(tname,'(4hTANG,i1)') npart
        setvar = palloc(npart,tname,nnu,2)
        call pzero(hr(np(npart)),nnu)

c       Free symbolic and numeric data
        call umf4fsym(symbolic)
        call umf4fnum(numeric )

c       Fill control with default values

        control(1) = 2
        call umf4def(control)

c     Solution steps for assembled equations

      else

c       Factor equations

        if(flags(2)) then

c         Adjust pointer array for C-values

          setvar = palloc(227,'USOL3',neq+1,1)
          mr(np(227)) = 0
          do i = 1,neq
            mr(np(227)+i) = mr(np(225)+i-1)
          end do ! i
          do i = 0,mr(np(227)+neq)
            mr(np(226)+i) = mr(np(226)+i) - 1
          end do ! i

c         Pre-order and symbolic analysis

          call umf4sym(neq,neq, mr(np(227)),mr(np(226)), hr(np(npart)),
     &                 symbolic, control, info)


          if(debug) then
            write(*,*) ' SYMBOLIC: NEQ =',neq
            write(*,2000) info(1),info(16), (info(21)*info(4))/2**20,
     &                                      (info(22)*info(4))/2**20,
     &                    info(23),info(24),info(25)
          endif

          if (info(1) .lt. 0) then
            write(*,*) 'Error occurred in umf4sym: ',info(1)
            call plstop()
          end if

c         Factor step

          call umf4num ( mr(np(227)), mr(np(226)), hr(np(npart)),
     &                 symbolic, numeric, control, info)

          if(debug) then
            write(*,2001) info(1), info(66), (info(41)*info(4))/2**20,
     &                                       (info(41)*info(4))/2**20,
     &                    info(43), info(44), info(45)
          endif

          if (info(1) .lt. 0) then
            write(*,*) 'Error occurred in umf4num: ',info(1)
            call plstop()
          end if

        endif

c       Perform solve

        if(flags(4)) then

          setvar = palloc(228,'USOL4',neq  ,2)
          do i = 1,neq
            hr(np(228)+i-1) = b(i)
          end do ! i

c         Solve system using LU factorization.

          sys = 0
          call umf4sol (sys, b, hr(np(228)), numeric, control, info)

          if (info(1) .lt. 0) then
            write(*,*) 'Error occurred in umf4sol: ',info(1)
            call plstop()
          end if

c         Compute 'energy' for convergence test

          aengy  = dot(b,hr(np(228)),neq)
          setvar = palloc(228,'USOL4',    0,2)

        endif

c       Purge storage in 'factor'

        if(flags(5)) then
          setvar = palloc(227,'USOL3',    0,1)
          setvar = palloc(226,'USOL2',    0,1)  ! allocated in uiters
          setvar = palloc(225,'USOL1',    0,1)  ! allocated in uiters
          call umf4fsym(symbolic)
          call umf4fnum(numeric )
        endif

      endif

c     Formats

2000  format(' Symbolic Analysis:',/,
     &       '   Status:  ',f5.0/
     &       '   Time:    ',e10.2, 'sec',/,
     &       '   Estimates (upper bound) for numeric LU:',/,
     &       '   Size of LU:     ',f10.2, 'MB',/,
     &       '   Memory needed:  ',f10.2, 'MB',/,
     &       '   Flop count:     ',e10.2,/,
     &       '   NNZ (L):        ',f10.0,/,
     &       '   NNZ (U):        ',f10.0)

2001  format(' Numeric factorization:',/,
     &       '   Status:  ',f5.0/
     &       '   Time:    ',e10.2, 'sec',/,
     &       '   Actual numeric LU statistics:',/,
     &       '   Size of LU:     ',f10.2, 'MB',/,
     &       '   Memory needed:  ',f10.2, 'MB',/,
     &       '   Flop count:     ',e10.2,/,
     &       '   NNZ (L):        ',f10.0,/,
     &       '   NNZ (U):        ',f10.0)

      end
